import tensorflow as tf

import numpy as np


def conv2d(x, W, stride=1):
    """conv2d returns a 2d convolution layer with stride."""
    return tf.nn.conv2d(x, W, strides=[1, stride, stride, 1], padding='SAME')


def max_pool_2x2(x):
    """max_pool_2x2 downsamples a feature map by 2X."""
    return tf.nn.max_pool(x, ksize=[1, 2, 2, 1], strides=[1, 2, 2, 1], padding='SAME')


def avg_pool_2x2(x):
    """avg_pool_2x2 downsamples a feature map by 2X."""
    return tf.nn.avg_pool(x, ksize=[1, 2, 2, 1], strides=[1, 2, 2, 1], padding='SAME')


def weight_variable(shape, trainable=True, stddev=0.1):
    """weight_variable generates a weight variable of a given shape."""
    return tf.get_variable('weights', shape=shape,
        trainable=trainable, initializer=tf.truncated_normal_initializer(stddev=stddev))


def bias_variable(shape, trainable=True, const=0.1):
    """bias_variable generates a bias variable of a given shape."""
    return tf.get_variable('biases', shape=shape,
        trainable=trainable, initializer=tf.constant_initializer(const))


class Convnet():
    def __init__(self, name):
        self.name = name
        self.load_vars = []
        self.train_vars = []
        self.conv_means = {}

    def convolutionLayer(self, layer_name, input_image, weights, stride=1, summary=False, load=False, train_loaded=False):
        trainable = (not load) or train_loaded
        with tf.variable_scope(layer_name):
            W_conv = weight_variable(weights, trainable=trainable)
            b_conv = bias_variable([weights[-1]], trainable=trainable)
            h_conv = tf.nn.relu(conv2d(input_image, W_conv, stride) + b_conv, name=layer_name)

            if load: self.load_vars += [W_conv, b_conv]
            if trainable: self.train_vars += [W_conv, b_conv]

            self.conv_means[layer_name] = tf.reduce_mean(h_conv, axis=-1, keepdims=True)
            if summary:
                tf.summary.image('activations', self.conv_means[layer_name], max_outputs=1)
                tf.summary.histogram('weights', W_conv)
            return h_conv


    def pooling2x2Layer(self, layer_name, input_image, ty='max'):
        with tf.variable_scope(layer_name):
            if ty == 'max':
                h_pool = max_pool_2x2(input_image)
            elif ty == 'avg':
                h_pool = avg_pool_2x2(input_image)
            else:
                raise
            return h_pool


    def fullyConnectedLayer(self, layer_name, input_vec, weights, summary=False, load=False, stddev=1.0, bias_const=0.1, train_loaded=False):
        trainable = (not load) or train_loaded
        with tf.variable_scope(layer_name):
            W_fc = weight_variable(weights, trainable=trainable, stddev=stddev)
            b_fc = bias_variable([weights[-1]], trainable=trainable, const=bias_const)
            h_fc = tf.nn.relu(tf.matmul(input_vec, W_fc) + b_fc, name=layer_name)

            if load: self.load_vars += [W_fc, b_fc]
            if trainable: self.train_vars += [W_fc, b_fc]

            if summary:
                tf.summary.histogram('activations', h_fc)
            return h_fc


    @property
    def all_vars(self):
        return list(set(self.train_vars + self.load_vars))


class ConvnetRegression(Convnet):

    def __init__(self, images):
        """convnet builds the graph for a convNet for regression.

        Args:
            images: shape (batch_size, 480, 480, 3)
        """
        super().__init__('cloud_regress')
        convolutionLayer = self.convolutionLayer
        pooling2x2Layer = self.pooling2x2Layer
        fullyConnectedLayer = self.fullyConnectedLayer

        # 480x480
        x = convolutionLayer('conv1', images, [7, 7, 3, 8], 2, summary=True)
        # 240x240
        x = pooling2x2Layer('pool1', x, 'max')
        # 120x120  
        x = convolutionLayer('conv2', x, [5, 5, 8, 16])
        x = convolutionLayer('conv3', x, [5, 5, 16, 16])
        x = convolutionLayer('conv4', x, [5, 5, 16, 16], 2, summary=True)
        # 60x60
        x = convolutionLayer('conv5', x, [5, 5, 16, 16])
        x = convolutionLayer('conv6', x, [5, 5, 16, 16])
        x = convolutionLayer('conv7', x, [5, 5, 16, 16], 2,summary=True)
        # 30x30
        x = pooling2x2Layer('pool2', x, 'avg')
        # 15x15
        x = tf.reshape(x, [-1, 15*15*16])
        # 15*15*16 = 3600
        x = fullyConnectedLayer('fc1', x, [15*15*16, 1024], summary=True)

        # Dropout - controls the complexity of the model, prevents co-adaptation of
        # features.
        with tf.name_scope('dropout'):
            keep_prob = tf.placeholder(tf.float32)
            x = tf.nn.dropout(x, keep_prob)

        x = fullyConnectedLayer('fc2', x, [1024, 1])

        self.fc2 = x
        self.keep_prob = keep_prob

    def loss(self, labels):
        with tf.name_scope('loss'):
            absLosses = tf.abs(labels - self.fc2)
            absLoss = tf.reduce_mean(absLosses)
            tf.summary.scalar('absLoss', absLoss)
        return absLoss

class ConvnetVGG16(Convnet):
    """
    Model by:
    Karen Simonyan & Andrew Zisserman (VGG)
    Very Deep Convolutional Networks For Large-Scale Image Recognition
    https://arxiv.org/pdf/1409.1556.pdf
    """
    def __init__(self, images, classes, height_regression=False):
        """Builds a graph with pretrained vgg16 convnet and custom
        FC-layers for cloud height classification.

        Args:
            images: shape (batch_size, 224, 224, 3)
            classes: count of classification categories
        """
        super().__init__('cloud_vgg16')
        self.num_height_classes = classes
        convolutionLayer = self.convolutionLayer
        pooling2x2Layer = self.pooling2x2Layer
        fullyConnectedLayer = self.fullyConnectedLayer

        # zero-mean input
        with tf.name_scope('preprocess') as scope:
            mean = tf.constant([123.68, 116.779, 103.939], dtype=tf.float32, shape=[1, 1, 1, 3], name='img_mean')
            images_m = images-mean

        # 224x224x3
        x = convolutionLayer('conv1_1', images_m, [3, 3, 3, 64], load=True)
        x = convolutionLayer('conv1_2', x, [3, 3, 64, 64], load=True)
        x = pooling2x2Layer('pool1', x, 'max')
        # 112x112x64
        x = convolutionLayer('conv2_1', x, [3, 3, 64, 128], load=True)
        x = convolutionLayer('conv2_2', x, [3, 3, 128, 128], load=True)
        x = pooling2x2Layer('pool2', x, 'max')
        # 56x56x128
        x = convolutionLayer('conv3_1', x, [3, 3, 128, 256], load=True, train_loaded=True)
        x = convolutionLayer('conv3_2', x, [3, 3, 256, 256], load=True, train_loaded=True)
        x = convolutionLayer('conv3_3', x, [3, 3, 256, 256], summary=True, load=True, train_loaded=True)
        x = pooling2x2Layer('pool3', x, 'max')
        # 28x28x256
        x = convolutionLayer('conv4_1', x, [3, 3, 256, 512], load=True, train_loaded=True)
        x = convolutionLayer('conv4_2', x, [3, 3, 512, 512], load=True, train_loaded=True)
        x = convolutionLayer('conv4_3', x, [3, 3, 512, 512], summary=True, load=True, train_loaded=True)
        x = pooling2x2Layer('pool4', x, 'max')
        # 14x14x512
        x = convolutionLayer('conv5_1', x, [3, 3, 512, 512], load=True, train_loaded=True)
        x = convolutionLayer('conv5_2', x, [3, 3, 512, 512], load=True, train_loaded=True)
        x = convolutionLayer('conv5_3', x, [3, 3, 512, 512], summary=True, load=True, train_loaded=True)
        x = pooling2x2Layer('pool5', x, 'max')
        # 7x7x512
        self.pool5 = x

        flat_length = int(np.prod(x.get_shape()[1:]))
        x = tf.reshape(x, [-1, flat_length])

        x = fullyConnectedLayer('fc6', x, [flat_length, 4096], summary=True, load=True, train_loaded=True,
                                stddev=0.1, bias_const=0.1)
        self.fc6 = x

        self.keep_prob = tf.placeholder(tf.float32)
        with tf.name_scope('dropout1'):
            x = tf.nn.dropout(x, self.keep_prob)

        x = fullyConnectedLayer('fc7', x, [4096, 4096], summary=True, load=True, train_loaded=True,
                                stddev=0.1, bias_const=0.1)
        self.fc7 = x
        with tf.name_scope('dropout2'):
            x = tf.nn.dropout(x, self.keep_prob)
        self.fc7_dropout = x

        with tf.variable_scope('fc8'):
            W_fc = weight_variable([4096, classes], stddev=0.1)
            b_fc = bias_variable([classes], const=0.1)
            h_fc = tf.matmul(x, W_fc) + b_fc

            self.train_vars += [W_fc, b_fc]

            tf.summary.histogram('activations', h_fc)

        self.fc8 = h_fc
        self.probs = tf.nn.softmax(self.fc8)

        if height_regression:
            self.height_reg_log = self.height_regression()

    def height_regression(self):
        """ height_reg_log = log(height_prediction + 1)"""
        with tf.variable_scope('height_regression_fc'):
            W_fc = weight_variable([4096, 1], stddev=0.1)
            b_fc = bias_variable([1], const=0.1)
            h_fc = tf.matmul(self.fc7_dropout, W_fc) + b_fc

            self.train_vars += [W_fc, b_fc]
            tf.summary.histogram('activations', h_fc)
        return tf.reshape(h_fc, [-1])


    # loss functions

    def cross_entropy_loss(self, labels):
        with tf.name_scope('cross_entropy_loss'):
            logits = self.fc8
            cross_entropy = tf.nn.sparse_softmax_cross_entropy_with_logits(\
                labels=labels, logits=logits)

            tf.summary.scalar('cross_entropy', tf.reduce_mean(cross_entropy))

        return cross_entropy


    def height_regression_loss(self, cloud_heights):
        """ Distance of finite heights (log scale) or 0 for infinite heights """
        with tf.name_scope('height_regression_loss'):
            height_loss = tf.abs(self.height_reg_log - tf.log(cloud_heights + 1))
            zeros = tf.zeros(tf.shape(height_loss))
            reg_loss = tf.where(tf.is_inf(cloud_heights), zeros, height_loss)

            tf.summary.scalar('height_log_error', tf.reduce_mean(reg_loss))
        return reg_loss

    def loss_classes_with_inf_and_reg(self, images, labels, height_readings):
        # height bin classification + infinity height classification
        classification_loss = self.cross_entropy_loss(labels)
        # regression loss (if height finite)
        height_regression_loss = self.height_regression_loss(height_readings)

        with tf.name_scope('loss'):
            # join losses from different tasks
            joined_loss = classification_loss + height_regression_loss
            tf.summary.scalar('joined_loss_mean', tf.reduce_mean(joined_loss_mean))

        return joined_loss

    # predictions

    def class_predictions(self):
        return tf.argmax(self.probs, 1)


    def height_predictions(self):
        return tf.exp(self.height_reg_log) - 1


    def accuracy(self, labels):
        with tf.name_scope('accuracy'):
            prediction = self.class_predictions()
            one_equals = tf.to_float(tf.equal(prediction, labels))
            accuracy = tf.reduce_mean(one_equals)
            tf.summary.scalar('batch_accuracy', accuracy)
        return accuracy

    # loading weights

    def var_loading_name(self, var_name):
        """Convert variable name to loadable weight name"""
        wnames = {'biases':'b', 'weights':'W'}
        variable_comps = var_name.split(':')[0].split('/')
        if len(variable_comps) == 2:
            weight_name = variable_comps[0] + '_' + wnames[variable_comps[1]]
            return weight_name


    def load_weights(self, weight_file, sess):
        """Load pretrained weights of the vgg16 model from .npz file

        Args:
            weight_file: .npz file with vgg16 weights
        """
        with np.load(weight_file) as weights:

            for i, param in enumerate(self.load_vars):
                weight_name = self.var_loading_name(param.name)
                if weight_name in weights:
                    print(i, weight_name, param.shape)
                    sess.run(param.assign(weights[weight_name]))
                else:
                    print(i, param, 'not in', weight_file, 'as', weight_name)

    ### multi variable regression

    def add_regression(self, dims):
        with tf.variable_scope('regression_fc'):
            W_fc = weight_variable([4096, dims], stddev=0.1)
            b_fc = bias_variable([dims], const=0.1)
            h_fc = tf.matmul(self.fc7_dropout, W_fc) + b_fc

            self.train_vars += [W_fc, b_fc]
            tf.summary.histogram('activations', h_fc)
        self.reg_fc = h_fc


    def sum_regression_loss(self, cloud_heights, rain_quantities, sun_insolations):
        with tf.name_scope('regression_loss'):
            # height: 150-7000; rain: 0-1.1; sun: 0-1000
            # weights:    1.0         2000.0      2.0
            weights = tf.constant([1.0, 2000.0, 2.0], dtype=tf.float32)
            targets = tf.stack([cloud_heights, rain_quantities, sun_insolations], axis=1)
            sum_abs_losses = tf.reduce_sum(tf.abs(self.reg_fc - targets) * weights, axis=1)

            tf.summary.scalar('sum_abs_loss_mean', tf.reduce_mean(sum_abs_losses))
        return sum_abs_losses


    def regs_height_predictions(self):
        return tf.gather(self.reg_fc, 0, axis=1)


    def regs_rain_predictions(self):
        return tf.gather(self.reg_fc, 1, axis=1)


    def regs_sun_predictions(self):
        return tf.gather(self.reg_fc, 2, axis=1)
