

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import argparse
import sys
import os
import tempfile
import time

import tensorflow as tf

import numpy as np
import matplotlib.pyplot as plt

from math import ceil
FLAGS = None


def deepnn(images):
  """deepnn builds the graph for a deep net for classifying digits.
  Args:
    images: shape (batch_size, 480, 480, 3)
  Returns:
    A tuple (height_pred, keep_prob). height_pred is a tensor of shape (N_examples)
  """


  # First convolutional layer - maps one grayscale image to 32 feature maps.
  with tf.name_scope('conv1'):
    W_conv1 = weight_variable([5, 5, 3, 16])
    b_conv1 = bias_variable([16])
    h_conv1 = tf.nn.relu(conv2d(images, W_conv1) + b_conv1)
    tf.summary.histogram('activations', h_conv1)

  # Pooling layer - downsamples by 2X. 480x480x16 -> 240x240x16
  with tf.name_scope('pool1'):
    h_pool1 = max_pool_2x2(h_conv1)

  # Second convolutional layer -- maps 16 feature maps to 32.
  with tf.name_scope('conv2'):
    W_conv2 = weight_variable([5, 5, 16, 32])
    b_conv2 = bias_variable([32])
    h_conv2 = tf.nn.relu(conv2d(h_pool1, W_conv2) + b_conv2)
    tf.summary.histogram('activations', h_conv2)

  # Second pooling layer. 240x240x32 -> 120x120x32
  with tf.name_scope('pool2'):
    h_pool2 = max_pool_2x2(h_conv2)

  # Third convolutional layer -- maps 32 feature maps to 32.
  with tf.name_scope('conv3'):
    W_conv3 = weight_variable([5, 5, 32, 32])
    b_conv3 = bias_variable([32])
    h_conv3 = tf.nn.relu(conv2d(h_pool2, W_conv3) + b_conv3)
    tf.summary.histogram('activations', h_conv3)

  # Second pooling layer. 120x120x64 -> 60x60x64
  with tf.name_scope('pool3'):
    h_pool3 = max_pool_2x2(h_conv3)
  # Fully connected layer 1 -- after 3 rounds of downsampling, our 480x480 image
  # is down to 60x60x32 feature maps -- maps this to 1024 features.
  with tf.name_scope('fc1'):
    W_fc1 = weight_variable([60*60*32, 1024])
    b_fc1 = bias_variable([1024])

    h_pool3_flat = tf.reshape(h_pool3, [-1, 60*60*32])
    h_fc1 = tf.nn.relu(tf.matmul(h_pool3_flat, W_fc1) + b_fc1)
    tf.summary.histogram('activations', h_fc1)

  # Dropout - controls the complexity of the model, prevents co-adaptation of
  # features.
  with tf.name_scope('dropout'):
    keep_prob = tf.placeholder(tf.float32)
    h_fc1_drop = tf.nn.dropout(h_fc1, keep_prob)

  # full connected layer to single value
  with tf.name_scope('fc2'):
    W_fc2 = weight_variable([1024, 1])
    b_fc2 = bias_variable([1])

    y_conv = tf.matmul(h_fc1_drop, W_fc2) + b_fc2
  return y_conv, keep_prob


def conv2d(x, W):
  """conv2d returns a 2d convolution layer with full stride."""
  return tf.nn.conv2d(x, W, strides=[1, 1, 1, 1], padding='SAME')


def max_pool_2x2(x):
  """max_pool_2x2 downsamples a feature map by 2X."""
  return tf.nn.max_pool(x, ksize=[1, 2, 2, 1],
                        strides=[1, 2, 2, 1], padding='SAME')


def weight_variable(shape):
  """weight_variable generates a weight variable of a given shape."""
  initial = tf.truncated_normal(shape, stddev=0.1)
  return tf.Variable(initial)


def bias_variable(shape):
  """bias_variable generates a bias variable of a given shape."""
  initial = tf.constant(0.1, shape=shape)
  return tf.Variable(initial)


# Transforms a scalar string `example_proto` into a pair of a scalar string and
# a scalar integer, representing an image and its label, respectively.
def _parse_function(example_proto):
  features = {"image_encoded": tf.FixedLenFeature((), tf.string, default_value=""),
              "width": tf.FixedLenFeature((), tf.int64, default_value=480),
              "height":tf.FixedLenFeature((), tf.int64, default_value=480),
              "depth":tf.FixedLenFeature((), tf.int64, default_value=1),
              "label": tf.FixedLenFeature((), tf.int64, default_value=0)}
  parsed_features = tf.parse_single_example(example_proto, features)
  image = tf.image.decode_image(parsed_features['image_encoded'], 3)
  image = tf.to_float(image)
  labels = tf.to_float(parsed_features["label"])
  return image, labels

DATASET_DIR = 'output'
TRAIN_FILE = 'train.tfrecords'
VALIDATION_FILE = 'validation.tfrecords'

def main(_):

  #use dataset data in graph
  with tf.name_scope('input'):
    train_dataset = tf.data.TFRecordDataset(os.path.join(DATASET_DIR, TRAIN_FILE)).map(_parse_function).shuffle(buffer_size=500).repeat(FLAGS.num_epochs)
    train_dataset = train_dataset.batch(FLAGS.batch_size)
    training_iterator = train_dataset.make_one_shot_iterator()

    val_dataset = tf.data.TFRecordDataset(os.path.join(DATASET_DIR, VALIDATION_FILE)).map(_parse_function).shuffle(buffer_size=500)
    val_dataset = val_dataset.batch(FLAGS.batch_size)
    validation_iterator = val_dataset.make_initializable_iterator()

    handle = tf.placeholder(tf.string, shape=[])
    iterator = tf.data.Iterator.from_string_handle(handle, train_dataset.output_types, train_dataset.output_shapes)
    images, labels = iterator.get_next()

  labels = tf.reshape(labels, [-1, 1])

  # Build the graph for the deep net
  y_conv, keep_prob = deepnn(images)

  with tf.name_scope('loss'):
    absLoss = tf.abs(labels - y_conv)
  absLoss = tf.reduce_mean(absLoss)
  tf.summary.scalar('absLoss', absLoss)

  with tf.name_scope('adam_optimizer'):
    train_step = tf.train.AdamOptimizer(1e-4).minimize(absLoss)


  init_op = tf.group(tf.global_variables_initializer(), tf.local_variables_initializer())

  with tf.Session() as sess:


    training_handle = sess.run(training_iterator.string_handle())
    validation_handle = sess.run(validation_iterator.string_handle())

    sess.run([validation_iterator.initializer])
    lbs = sess.run([labels], feed_dict={keep_prob: 0.5, handle: validation_handle})
    print(lbs)

    merged = tf.summary.merge_all()
    graph_location = os.path.join(FLAGS.summaries_dir, 'train')
    print('Saving graph to: %s' % graph_location)
    train_writer = tf.summary.FileWriter(graph_location, sess.graph)
    test_writer = tf.summary.FileWriter(os.path.join(FLAGS.summaries_dir, 'test'))

    

    sess.run(init_op)

    try:
      step = 0
      while True:
        start_time = time.time()

        summary, _ = sess.run([merged, train_step], feed_dict={keep_prob: 0.5, handle: training_handle})
        if step % 2 == 0:
          train_writer.add_summary(summary, step)

        if step % 2 == 1:
          #reinitialize the validation_iterator
          sess.run([validation_iterator.initializer])
          summary, train_conv, train_lbs, train_absLoss = sess.run([merged, y_conv, labels, absLoss], feed_dict={keep_prob: 1.0, handle: validation_handle})
          duration = time.time() - start_time
          print('step %d, training loss %g (%d ms)' % (step, train_absLoss, duration))
          print(train_conv, train_lbs)

          test_writer.add_summary(summary, step)

        step += 1
        if step == 10:
          break
    except tf.errors.OutOfRangeError:
      print('Done training for %d epochs, %d steps.' % (FLAGS.num_epochs, step))

    
    
if __name__ == '__main__':
  parser = argparse.ArgumentParser()
  parser.add_argument('--num_epochs', type=int,
                    default=1,
                    help='Number of epochs for training')
  parser.add_argument('--batch_size', type=int,
                    default=3,
                    help='Batch size for learning')
  parser.add_argument('--summaries_dir', type=str,
                      default='summaries',
                      help='Directory for storing summary')
  FLAGS, unparsed = parser.parse_known_args()
  tf.app.run(main=main, argv=[sys.argv[0]] + unparsed)


